from p2g import scalar


# Just type hints, can't be called.
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name


def abs(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


# def print(_v) -> scalar.ConstantNone:  # no cover
#     return scalar.ConstantNone()


def acos(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def asin(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def atan(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def cos(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def exp(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def fix(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def fup(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def ln(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def sin(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def sqrt(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def tan(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def exists(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def ground(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()
